#include <iostream>
#include <fstream>
#include <string>
#include <vector>

#include <windows.h>
#include <libloaderapi.h>

#include "oodle2.h"

const LPCWSTR LIB_NAME = L"oo2core_8_win64";

void usage() {
    std::cerr <<
        "Usage: oodle-cli [OPTIONS] <in_file> <out_file> <uncompressed_size>\n"
        "Decompress a <in_file> to <out_file> using the Oodle algorithm.\n"
        "The size of the uncompressed data must be known.\n"
        "\n"
        "  -v Increase verbosity. May be specified up to three times.\n";
}

int main(int argc, char* argv[])
{
    HINSTANCE hDLL;
    hDLL = LoadLibrary(LIB_NAME);
    if (hDLL == NULL) {
        std::cerr << "ERROR: Couldn't find library file '" << LIB_NAME << ".dll'!\n";
        return 1;
    }

    auto decompress = (decltype(OodleLZ_Decompress)*)GetProcAddress((HMODULE)hDLL, "OodleLZ_Decompress");
    if (decompress == NULL) {
        std::cerr << "ERROR: The library is incompatible!\n";
        return 1;
    }
    int verbosity = 0;

    size_t i = 1;
    for (; i < argc; i++)
    {
        if (strcmp(argv[i], "-v") == 0) {
            if (verbosity < 3) {
                verbosity++;
            }
        }
        else {
            break;
        }
    }

    if (argc - i < 3) {
        std::cerr << "Arguments missing!\n\n";
        usage();
        return 1;
    }

    std::string in_name = argv[i];
    std::string out_name = argv[i + 1];
    size_t raw_buffer_size = std::stoi(argv[i + 2]);

    std::cout << "Attempting to decompress from '" << in_name << "' to '" << out_name << "' (" << raw_buffer_size << " bytes), with verbosity " << verbosity << ".\n";

    std::ifstream in_file(in_name, std::ios::binary | std::ios::ate);
    if (!in_file) {
        std::cerr << "Failed to open compressed file!\n";
        return 1;
    }

    std::ofstream out_file(out_name, std::ios::binary);
    if (!out_file) {
        std::cerr << "Failed to open output file!\n";
        return 1;
    }

    std::streamsize compressed_buffer_size = in_file.tellg();
    in_file.seekg(0, std::ios::beg);

    std::vector<char> compressed_buffer(compressed_buffer_size);
    in_file.read(compressed_buffer.data(), compressed_buffer_size);

    if (in_file.fail()) {
        std::cerr << "Failed to read compressed file!\n";
        return 1;
    }

    std::vector<char> raw_buffer(raw_buffer_size);

    int res = decompress(
        compressed_buffer.data(),
        compressed_buffer_size,
        raw_buffer.data(),
        raw_buffer_size,
        OodleLZ_FuzzSafe_Yes,
        OodleLZ_CheckCRC_No,
        (OodleLZ_Verbosity)verbosity,
        nullptr,
        0,
        nullptr,
        nullptr,
        nullptr,
        0,
        OodleLZ_Decode_Unthreaded
    );

    out_file.write(raw_buffer.data(), res);

    if (out_file.fail()) {
        std::cerr << "Failed to write output file!\n";
        return 1;
    }

    if (res != raw_buffer_size) {
        std::cerr << "Failed to decompress. Expected " << raw_buffer_size << " bytes, got " << res << "!\n";
        return 1;
    }

    std::cout << "Done!\n";
}
